#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <utime.h>
#include <errno.h>
#include <fcntl.h>

int copy_file(const char *src_file, const char *dest_file, int overwrite, int copy_attr)
{
    int     src_fd;
    int     dest_fd;
    struct  stat sts;
    char    data_buf[4096];
    int     tmp_errno;
    int     size;
    struct  utimbuf attr;
    
    if((src_fd = open(src_file, O_RDONLY)) == -1) {
        return -1;
    }

    /* 원본 file의 속성을 읽습니다. */
    fstat(src_fd, &sts);

    if(overwrite) { /* 이미 파일이 있으면 overwrite를 하겠다면... */
        dest_fd = open(dest_file, O_WRONLY | O_CREAT | O_TRUNC, sts.st_mode);
    } else {        /* 파일이 있으면, 생성하지 말라고 설정한 경우 */
        dest_fd = open(dest_file, O_WRONLY | O_CREAT | O_EXCL, sts.st_mode);
    }

    if(dest_fd == -1) {
        tmp_errno = errno; 
        close(src_fd);  
        errno = tmp_errno; // close가 초기화한 errno를 복구함
        return -1;
    }

    while(size = read(src_fd, data_buf, 4096)) {
        if(size == -1) {
            if(errno == EINTR) {
                continue;
            }
            tmp_errno = errno;
            close(src_fd);
            close(dest_fd);
            errno = tmp_errno; // close가 초기화한 errno를 복구함
            return -1;
        }

        while(write(dest_fd, data_buf, size) == -1) {
            if(errno == EINTR) {
                /* signal이 발생한 경우에는 재작업 */
                continue;
            } else {
                /* disk가 full났거나 무슨 일이 있음. */
                tmp_errno = errno;
                close(src_fd);
                close(dest_fd);
                errno = tmp_errno; // close가 초기화한 errno를 복구함
            }
        }
    }
    close(src_fd);
    close(dest_fd);
    
    /* 원본 파일의 속성을 복원해야 한다면... */
    if(copy_attr) {
        /* last access 시간, last modify 시간 복구 */
        attr.actime  = sts.st_atime;
        attr.modtime = sts.st_mtime;
        utime(dest_file, &attr);  
        
        /* 원본 파일의 파일 권한을 복원하기 
        * open시에 파일권한을 설정하였지만, 
        * 이미 존재했던 파일은 파일권한이 기존 파일의 권한이므로
        * 파일의 권한도 복구합니다.
        */
        chmod(dest_file, sts.st_mode);
    }
    return 0;
}